#include <ESP8266WiFi.h>
#include <PubSubClient.h>
#include <WiFiClient.h>

/**
 *  Change wifi and mqtt server parameters
 */
//#define WIFI_SSID "ssid"
//#define WIFI_PASS "pass"
//#define MQTT_BROKER "homeassistant.local"
//#define MQTT_USER "mqtt-user"
//#define MQTT_PASS "mqtt-pass"

#define ALARM_PAYLOAD_CONFIG "{\"avty_t\":\"%s\",\"cmd_t\":\"%s\",\"name\":\"%s\",\"pl_on\":\"%s\"}"
#define PIN_BUZZER D2
#define TONE_KHZ 1100
#define PITCH_LOW 200
#define PITCH_HIGH 1000
#define PANIC_ALARM_DURATION 15000

/**
 * Homeassistant MQTT Discovery topics
 */
char alarm_topic_cmd[] = "homeassistant/scene/esp8266/cmd";
char alarm_topic_available[] = "homeassistant/scene/esp8266/available";
char alarm_topic_config_armed[] = "homeassistant/scene/esp8266-armed/config";
char alarm_topic_config_disarmed[] = "homeassistant/scene/esp8266-disarmed/config";
char alarm_topic_config_panic[] = "homeassistant/scene/esp8266-panic/config";
bool flag_armed = false;
bool flag_disarmed = false;
bool flag_panic = false;
int pitchStep = 10;
int currentPitch = PITCH_LOW;
unsigned long panicMillis = 0;

void callback(char* topic, byte* payload, unsigned int length);
void reconnect_mqtt();

WiFiClient client;
PubSubClient mqtt(MQTT_BROKER, 1883, callback, client);

/**
 * Connect to WiFi and MQTT
 */
void setup() {
  pinMode(PIN_BUZZER, OUTPUT);

  WiFi.begin(WIFI_SSID, WIFI_PASS);
  if (WiFi.waitForConnectResult() != WL_CONNECTED) {
    ESP.restart();
  }
  
  reconnect_mqtt();
}

/**
 * Keep MQTT connection alive,
 * if some flag is set, play tone
 */
void loop() {
  if (!mqtt.loop()) {
    reconnect_mqtt();  
  }

  if (flag_armed) {
    flag_armed = false;
    tone(PIN_BUZZER, TONE_KHZ);
    delay(100);
    noTone(PIN_BUZZER);
  } else if (flag_disarmed) {
    flag_disarmed = false;
    tone(PIN_BUZZER, TONE_KHZ);
    delay(100);
    noTone(PIN_BUZZER);
    delay(125);
    tone(PIN_BUZZER, TONE_KHZ);
    delay(100);
    noTone(PIN_BUZZER);
  } else if (flag_panic) {
    tone(PIN_BUZZER, currentPitch, 10);
    currentPitch += pitchStep;
    if(currentPitch >= PITCH_HIGH) {
      pitchStep = -pitchStep;
    } else if(currentPitch <= PITCH_LOW) {
      pitchStep = -pitchStep;
    }
    delay(10);
    if (millis() - panicMillis >= PANIC_ALARM_DURATION) {
      flag_panic = false;
    }
  }
}

/**
 * Connect to mqtt broker,
 * set device availability as Last Will Testament,
 * publish Hoemassistant MQTT Discovery messages
 */
void reconnect_mqtt() {
  char payload[255];

  while (!mqtt.connected()) {
    if (mqtt.connect("esp8266", MQTT_USER, MQTT_PASS, alarm_topic_available, 0, true, "offline")) {
      mqtt.subscribe(alarm_topic_cmd);
      snprintf(payload, sizeof(payload), ALARM_PAYLOAD_CONFIG, alarm_topic_available, alarm_topic_cmd, "Alarm armed", "ARMED");
      mqtt.publish(alarm_topic_config_armed, payload, true);
      delay(1000);
      snprintf(payload, sizeof(payload), ALARM_PAYLOAD_CONFIG, alarm_topic_available, alarm_topic_cmd, "Alarm disarmed", "DISARMED");
      mqtt.publish(alarm_topic_config_disarmed, payload, true);
      delay(1000);
      snprintf(payload, sizeof(payload), ALARM_PAYLOAD_CONFIG, alarm_topic_available, alarm_topic_cmd, "Alarm panic", "PANIC");
      mqtt.publish(alarm_topic_config_panic, payload, true);
      delay(1000);
      mqtt.publish(alarm_topic_available, "online", true);
    } else {
      delay(5000);
    }
  }
}

/**
 * Message from broker is rceived
 * only set flag, don't call any function
 */
void callback(char* topic, byte* payload, unsigned int length) {
  if (strncmp("ARMED", (char*)payload, length) == 0) {
    flag_panic = false;
    flag_armed = true;
  } else if (strncmp("DISARMED", (char*)payload, length) == 0) {
    flag_panic = false;
    flag_disarmed = true;
  } else if (strncmp("PANIC", (char*)payload, length) == 0) {
    panicMillis = millis();
    flag_panic = true;
  }
}
